__all__ = ['Playlist', 'EndOfPlaylist']

import random # shuffle
import web
from zicbee.core.httpdb import web, WEB_FIELDS
from zicbee_lib.debug import DEBUG
from zicbee_lib.formats import compact_int

class EndOfPlaylist(Exception): pass

class Playlist(list):
    """ A playlist class
    """

    def __init__(self, *args):
        list.__init__(self, *args)
        self.pos = -1
        self.sense = 1

    def __delitem__(self, slc):
        """ Deletes an entry (int and slices are supported) """
        self.__update_position(slc)
        try:
            idx = int(slc)
        except TypeError:
            # slice
            idx = slc.start

        if idx < self.pos:
            self.pos -= 1

        return list.__delitem__(self, slc)

    def __setitem__(self, slc, val):
        self.__update_position(slc)
        return list.__setitem__(self, slc, val)

    def __update_position(self, start, stop=None):

        if stop is None:
            try:
                start = int(start)
            except TypeError:
                try:
                    stop = start[1]
                    start = start[0]
                except TypeError:
                    # slice
                    stop = start.stop
                    start = start.start
            else:
                stop = start

        if start < self.pos:
            # increment
            self.pos += 1
        else:
            if start < self.pos and stop < self.pos:
                # increment stop-start
                self.pos += stop-start
            elif start < self.pos < stop:
                # reset position to beginning of slice
                self.pos = 0 if self.playlist else -1

    def insert(self, idx, obj):
        """ Inserts an object at a position, behaves like `list.insert`.
        """
        self.__update_position(idx)
        list.insert(self, idx, obj)

    def inject(self, data, position=None):
        """ inject a song or a range of songs,
        keeps the position if possible """
        if hasattr(data, '__getitem__') and isinstance(data[0], basestring):
            data = [data]
        elif isinstance(data, slice):
            data = self[data]

        p = self.pos+1 if position is None else position

        self.__update_position(p)

        self[p:p] = data

    def extend(self, data, idx=None):
        """ See `list.extend` """
        if idx is None:
            idx = len(self)
        self.__update_position(idx, idx+len(data))
        list.extend(self, data)

    def replace(self, any_iterable):
        current = self.selected
        if current:
            self.pop(self.pos)
        self[:] = any_iterable
        if current:
            self.insert(0, current)
            self.pos = 0
        else:
            self.pos = -1

    def shuffle(self):
        if len(self) == 0:
            return
        current = self.selected
        if current:
            self.pop(self.pos)
        random.shuffle(self)
        if current:
            self.insert(0, current)
            self.pos = 0
        else:
            self.pos = -1

    def clear(self):
        self[:] = []
        self.pos = -1

    def _selected(self, p=None):
        if not p:
            p = self.pos

        if p == -1 or len(self) == 0:
            return None
        try:
            return self[p]
        except IndexError:
            DEBUG()

    selected = property(_selected)

    @property
    def sibling(self):
        return self._selected_dict(self.pos+self.sense)

    def _selected_dict(self, pos=None):
        sel = self._selected(pos)

        web.debug('SELECTED: %r'%sel)
        if not sel:
            return dict()

        d = dict(zip(['uri']+WEB_FIELDS, sel))

        try:
            d['length'] = int(d.get('length') or 0)
            d['score'] = int(d.get('score') or 0)
            d['tags'] = d.get('tags') or u''
            d['pls_position'] = self.pos
            d['pls_size'] = len(self)
        except ValueError, e:
            # !!corrupted data!!
            web.debug('data corruption: %s'%e)
            return dict()
        except Exception:
            DEBUG()
        try:
            d['id'] = compact_int(d.pop('__id__'))
        except KeyError:
            pass
        return d

    selected_dict = property(_selected_dict)

    def move(self, steps):
        self.pos += steps
        self.sense = steps
        if self.pos >= len(self):
            self.pos = -1
            raise EndOfPlaylist()
        elif self.pos < -1:
            self.pos = -1




__all__ = ['Databae', 'valid_tags']
# vim: ts=4 sw=4 et

import os
import buzhug
from itertools import chain
from zicbee_lib.config import DB_DIR, media_config
import re
import time

try:
    required_buzhug = [1, 5]
    assert [int(d) for d in buzhug.version.split('.')] >= required_buzhug
except AssertionError:
    raise SystemExit('Wrong buzhug installed, please install at least version %s'%('.'.join(required_buzhug)))


valid_ext = media_config.keys()

valid_tags = (
        'genre',
        'artist',
        'album',
        'title',
        'track',
        'filename',
        'score',
        'tags',
        'length')

track_re = re.compile('.*?(\d+)')

filters_dict = dict(
        track = ('TRCK', 'tracknumber', u'WM/Track', "WM/TrackNumber" 'trkn'),
        title = ('Title', 'TITLE', 'TIT1', 'TIT2', '\xa9nam'),
        artist = ('Author', 'author', 'AUTHOR', 'TPE1', '\xa9ART', '\xa9aART', 'WM/Composer', 'WM/Writer', 'TCOM', 'TOPE', 'album artist'),
        album = (u'WM/AlbumTitle', 'TALB', 'debut album', '\xa9alb'),
        genre = ('\xa9gen', u'WM/Genre'),
        # disc = ('disknumber', "WM/PartOfSet"),
        # rating = ('Rating',),
        )


def checkdb(base_fn):
    return base_fn
    def _mkdec(somefn):
        def _auto_db_check(self, *args, **kw):
            if self.databases is not None:
                self._create()
            return somefn(self, *args, **kw)
        return _auto_db_check

    return _mkdec(base_fn)


class Database(object):

    DB_SCHEME = (
                ('filename', str),
                ('genre', unicode),
                ('artist', unicode),
                ('album', unicode),
                ('title', unicode),
                ('track', int),
                ('length', int),
                ('score', int),
                ('tags', unicode),
                )

    def __init__(self, name):
        """ Open/Create a database """
        self.databases = dict()
        self.db_name = name
        for name in name.split(','):
            p = os.path.join(DB_DIR, name)
            h = self._create(p)
            self.databases[name] = dict(
                    path = p,
                    handle = h,
                    )

    def add_error(self, message):
        f = getattr(self, '_error_fd', None)
        if not f:
            f = self._error_fd = open( os.path.join(DB_DIR, 'scan.log'), 'a+')
            self._error_fd.write("-- Errors scanning %s on %s --\n"%(self.db_name, time.asctime()))
        f.write(message+'\n')

    @property
    def artists(self):
        return set(r.artist for r in self.search(['artist']))

    @property
    def albums(self):
        return set(r.album for r in self.search(['album']))

    @property
    def genres(self):
        return set(r.genre for r in self.search(['genre']))

    def _create(self, db_name=None):
        if isinstance(db_name, basestring):
            databases = [ buzhug.Base(db_name) ]
        elif isinstance(db_name, buzhug.Base):
            databases = [ db_name ]
        else:
            for name, db in self._dbs_iter():
                self.databases[name] = buzhug.Base(db_name)
            databases = self.databases.values()

        kw = dict(mode='open')

        for db in databases:
            db.create(*self.DB_SCHEME, **kw)
        if databases is not self.databases and len(databases) == 1:
            return databases[0]
        return databases

    def _dbs_iter(self):
        return self.databases.iteritems()

    def close(self):
        for name, db in self._dbs_iter():
            db['handle'].close()

    def destroy(self):
        for name, db in self._dbs_iter():
            db['handle'].destroy()

    def commit(self):
        for name, db in self._dbs_iter():
            db['handle'].commit()

    def cleanup(self):
        for name, db in self._dbs_iter():
            db['handle'].cleanup()

    def update(self, record, **kw):
        for name, db in self._dbs_iter():
            db['handle'].update(record, **kw)

    @checkdb
    def __getitem__(self, item):
        for name, db in self._dbs_iter():
            db = db['handle']
            try:
                return db[item]
            except:
                continue

    @checkdb
    def __len__(self):
        return sum(len(db['handle']) for name, db in self._dbs_iter())

    @checkdb
    def search(self, *args, **kw):
        return chain( *(db['handle'].select(*args, **kw) for name, db in self._dbs_iter()) )

    @checkdb
    def u_search(self, *args, **kw):
        return chain( db['handle'].select_for_update(*args, **kw) for name, db in self._dbs_iter() )

    @checkdb
    def dump_archive(self, filename):
        """ dump a bz2 archive from this database """
        use_suffixes = len(self.databases) > 1
        self.cleanup()
        from tarfile import TarFile

        for name, db in self._dbs_iter():
            tar = TarFile.open('%s%s'%(filename, '_%s'%name if use_suffixes else ''), 'w:bz2')
            prefix_sz = len(db['path'])

            for root, dirs, files in os.walk(db['path']):
                short_root = root[prefix_sz:]
                for fname in files:
                    fd = file(os.path.join(root, fname))
                    ti = tar.gettarinfo(arcname=os.path.join(short_root, fname), fileobj=fd)
                    tar.addfile(ti, fileobj=fd)
            tar.close()

    def get_hash_iterator(self):
        """ Returns an (id, hash) tuple generator """
        for item in self.search():
            yield item.__id__, '%(artist)s:%(album)s:%(title)s'%dict(
                    (k, ''.join(c for c in getattr(item, k).lower() if c != ' '))
                    if isinstance(getattr(item, k), basestring)
                    else (k, getattr(item, k))
                    for k in item.fields)

    @checkdb
    def merge(self, db_name, directory=None, archive=None, no_dups=True, update=False):
        """ Merge informations from files in specified directory or archive """
        # TODO: handle no_dups to show merge of tags while adding new

        try:
            EasyID3
        except NameError:
            from mutagen import File
            from mutagen.mp3 import MP3
            from mutagen.easyid3 import EasyID3

        us_prefix = None

        print "Building data sets..."
        db_filenames = {}
        for name, db in self._dbs_iter():
            db = db['handle']
            db_filenames.update( (i.filename, i) for i in db if i.filename.startswith(directory) )

        db = self.databases[db_name]['handle']

        # Directory handling
        filenames = []
        if directory is not None and not us_prefix:
            for root, dirs, files in os.walk(directory):
                for fname in files:
                    if fname.rsplit('.', 1)[-1].lower() in valid_ext:
                        filenames.append( os.path.join(root, fname) )

        filenames = set(filenames)
        db_set = set(db_filenames.iterkeys())

        # remove old files
        print "\nRemoving..."
        for removed in db_set.difference(filenames):
            db.delete(db_filenames[removed])

        # add new files
        print "\nAdding..."
        for fullpath in filenames.difference(db_set):
            tags = None

            try:
                tags = File(fullpath)
            except Exception, e:
                print 'Error reading %s: %s'%(fullpath, e)
                yield "E"
                continue

            if tags:
                # Do it before tags is changed !
                length = int(tags.info.length+0.5)
                if isinstance(tags, MP3):
                    tags = EasyID3(fullpath)
                tags = dict(tags)
            else:
                length = 0

                name = unicode(fname, 'utf-8', 'replace')
                album = unicode(os.path.split(root)[-1], 'utf-8', 'replace')
                tags = {'title': name, 'album': album, 'artist': name}

            data = filter_dict(tags)
            data['filename'] = fullpath
            data['length'] = length
            if data.get('title') and data.get('artist'):
                yield '.'
            else:
                yield '0'
            try:
                db.insert(**data)
            except:
                import pdb; pdb.set_trace()

        if update:
            # Update tags
            print "\nUpdating..."
            for fullpath in filenames.intersection(db_set):
                tags = None

                try:
                    tags = File(fullpath)
                except Exception, e:
                    self.add_error("%s | %s | %r"%(db_name, fullpath, e))
                    yield "E"
                    continue

                if tags:
                    # Do it before tags is changed !
                    length = int(tags.info.length+0.5)
                    if isinstance(tags, MP3):
                        tags = EasyID3(fullpath)
                    tags = dict(tags)
                else:
                    length = 0

                    name = unicode(fname, 'utf-8', 'replace')
                    album = unicode(os.path.split(root)[-1], 'utf-8', 'replace')
                    tags = {'title': name, 'album': album, 'artist': name}

                data = filter_dict(tags)

                data['filename'] = fullpath
                data['length'] = length
                if data.get('title') and data.get('artist'):
                    yield '.'
                else:
                    yield '0'
                try:
                    song = db_filenames[fullpath]
                    #if any(data[k] != getattr(song, k) for k in data.keys()): # Looks useless
                    db.update(song, **data)
                except:
                    import pdb; pdb.set_trace()

        print "\nCleaning up..."
        self.cleanup()
        # Archive handling
        if  archive is not None:
            from tarfile import TarFile
            from tempfile import mkdtemp
            from shutil import rmtree
            tar = TarFile.open(archive)
            tmp = mkdtemp()
            try:
                # Extract archive to disk
                buf_size = 2**16
                for ti in tar.getmembers():
                    out_fname = os.path.join(tmp, ti.name)
                    in_fd = tar.extractfile(ti)
                    out_fd = file(out_fname, 'w')
                    while True:
                        data = in_fd.read(buf_size)
                        if not data:
                            break
                        out_fd.write(data)

                    out_fd.close()

                # Do the buzhug part
                tmp_db = self._create(tmp)

                for alien_entry in tmp_db:
                    if alien_entry not in db:
                        entry_dict = dict()
                        for f in alien_entry.fields:
                            if f[0] == '_':
                                continue
                            val = getattr(alien_entry, f)
                            if isinstance(val, str) and f != 'filename':
                                val = unicode(val)
                            entry_dict[f] = val
                        db.insert(**entry_dict)
                        yield '.'
            finally:
                rmtree(tmp, ignore_errors=True)

        db.commit()

def filter_dict(data):
    """ Returns a filtered given data dict """
    for good_tag, bad_tags in filters_dict.iteritems():
        for bad_tag in bad_tags:
            if good_tag in data:
                break
            if data.get(bad_tag) is not None:
                data[good_tag] = data[bad_tag]

    for k, v in data.items():
        if k not in valid_tags:
            del data[k]
        else:
            while isinstance(v, (list, tuple)):
                try:
                    v = v[0]
                except IndexError:
                    v = None
            data[k] = v

    for k in ('genre', 'artist', 'album', 'title'):
        d = data.get(k)
        if d is None:
            data[k] = u''
        elif not isinstance(d, unicode):
            data[k] = unicode(d)
        data[k] = data[k].split('\n')[0].replace(os.path.sep, '_').strip()

    if data.get('genre') == u'12':
        data['genre'] = u''


    track_val = data.get('track')
    if track_val is not None:
        try:
            track_val = unicode(track_val).strip()
        except Exception, e:
            print e
            track_val = unicode(track_val.value)
        try:
            if isinstance(track_val, (list, tuple)):
                track_val = track_val[0]
            if not isinstance(track_val, int):
                track_val = int( track_re.match(track_val).groups()[0] )
        except:
            print "Unable to get track for", repr(track_val)
            data['track'] = None
        else:
            try:
                data['track'] = int(track_val.strip())
            except:
                data['track'] = None
        if not data['title'] and data['track']:
            data['title'] = unicode(track_val)

    return data




# vim: et ts=4 sw=4
import itertools
import os
import sys
from time import time
from zicbee.core import zshell
from zicbee_lib.debug import DEBUG
from zicbee_lib.formats import duration_tidy, clean_path
from os.path import dirname

def _scan(update=False, **kw):
    newline_iterator = itertools.cycle(x == 20 for x in xrange(21))
    print ', '.join(':'.join((k,v)) for k,v in kw.iteritems())
    try:
        newlined=True
        for status_char in zshell.songs.merge(update=update,**kw):
            newlined=False
            print status_char,
            if newline_iterator.next():
                newlined=True
                print ''
            sys.stdout.flush()
        if not newlined:
            print ''
    except Exception, e:
        DEBUG()

def do_scan(up=False):
    """ Scan a directory for songs (fill Database)
    See "help" for a more complete documentation
    parameter:
      up: if True, updates tags of already scanned songs
    """
    if not zshell.args:
        sys.exit('At least one argument must be specified!')

    orig_nb = len(zshell.songs)
    start_t = time()

    archives = []
    directories = []

    for path in zshell.args:
        path = clean_path(path)
        if os.path.isdir(path):
            directories.append(path)
        else:
            archives.append(path)

    for path in archives:
        _scan(archive=path, db_name=zshell.songs.db_name)

    for path in directories:
        _scan(directory=path, db_name=zshell.songs.db_name, update=up)

    elapsed = time() - start_t
    delta = len(zshell.songs)-orig_nb
    print "\nProcessed %d (%s%d) songs in %s (%.2f/s.)"%(
            len(zshell.songs),
            '-' if delta < 0 else '+',
            abs(delta),
            duration_tidy(elapsed),
            len(zshell.songs)/elapsed)





from __future__ import with_statement

__all__ = ['PlayerCtl']

import os
import gc
import thread
import urllib
import atexit
import itertools
from functools import partial
from time import sleep
from Queue import Queue, Empty
from zicbee.utils import notify
from zicbee.core.httpdb import web
from threading import RLock, Thread, Event
from zicbee_lib.resources import get_players
from zicbee.core.playlist import EndOfPlaylist, Playlist
from zicbee_lib.config import config, media_config, DB_DIR
from zicbee_lib.debug import log, DEBUG
from zicbee_lib.formats import jload
from zicbee_lib.debug import nop
from zicbee_lib.parser import parse_string, PLAYLIST, tokens2string

try:
    from cPickle import Pickler, Unpickler
except ImportError:
    from Pickle import Pickler, Unpickler

from hashlib import md5

def get_type_from_uri(uri):
    if uri:
        try:
            return uri.rsplit('song.', 1)[1].split('?', 1)[0]
        except IndexError:
            return 'dunno'
            return uri.rsplit('.', 1)[1]
    return 'mp3'

def uri2fname(uri):
    return "%s.%s.%s"%(config.streaming_file, md5(uri).hexdigest(), str(get_type_from_uri(uri)))

class Downloader(Thread):
    q = Queue(3)
    next = None
    preloaded = []

    def __init__(self):
        Thread.__init__(self)
        self.aborted = False
        atexit.register(self.stop)

    def run(self):
        self.running = True
        stream = None
        fd = None
        cs = None

        while self.running:
            got_one = False
            while True:
                try:
                    uri, ic, cs, cb = self.q.get_nowait()
                    got_one = True
                except Empty:
                    break

            if got_one: # we have something to download
                if stream:
                    web.debug('Closing incomplete download %s'%preload_name)
                    fd.close()
                    stream.close()
                    self._abort(preload_name)
                preload_name = uri2fname(uri)
                self.preloaded.append(preload_name)
                fd = file(preload_name, 'w')
                stream = urllib.urlopen(uri)
                fd.write(stream.read(ic))
                fd.flush()
                cb()
            elif stream:
                d = stream.read(cs)
                if not d:
                    web.debug('download: EOF')
                    stream.close()
                    stream = None
                    fd.close()
                else:
                    fd.write(d)
            else:
                gc.collect()
                sleep(1) # wait some idle time
                if self.next: # preload asked, set the right descriptors
                    next = self.next
                    self.next = None
                    preload_name = uri2fname(next)
                    self.preloaded.append(preload_name)
                    fd = file(preload_name, 'w')
                    stream = urllib.urlopen(next)

    def _abort(self, name):
        web.debug("RM %s"% name)
        self.preloaded.remove(name)
        os.unlink(name)

    def get(self, uri, i_chunk=None, chunk=None):
        MAX_PRELOADS = 5
        filename = uri2fname(uri)
        web.debug('GET %s => %s'%(uri, filename))

        if len(self.preloaded) > MAX_PRELOADS:
            rm = self.preloaded.pop(0)
            try:
                web.debug("RM %s"% rm)
                os.unlink(rm)
            except OSError:
                pass
        if filename not in self.preloaded:
            # ask for a download and wait for the initial_chunk to be completed
            e = Event()
            self.q.put( (uri, i_chunk, chunk, e.set) )
            e.wait()
        return filename

    def stop(self):
        self.running = False
        for rm in self.preloaded:
            try:
                os.unlink(rm)
            except OSError:
                pass

class PlayerCtl(object):
    """ The player interface, this should lead to a constant code, with an interchangeable backend
    See documentation for the `zicbee.player`_ entry_point for the needed interface.
    """
    downloader = Downloader()

    def __init__(self):
        self.playlist = Playlist()
        self.downloader.start()
        self.views = []
        players = []
        preferences = [n.strip() for n in config.players]
        web.debug('player preferences: %s'%(', '.join(preferences)))
        for player_plugin in get_players():
            if player_plugin.name in preferences:
                players.insert(preferences.index(player_plugin.name), player_plugin)
            else:
                players.append(player_plugin)
        web.debug('available players: %s'%players)
        for player_plugin in players:
            try:
                self.player = player_plugin.load()()
            except Exception, e:
                web.debug('failed loading %s: %s'%(player_plugin.name, e))
            else:
                break # plugin loaded
        else:
            DEBUG()
            raise RuntimeError("No backend could be loaded!")

        self.position = None
        self._lock = RLock()
        self._paused = False
        thread.start_new_thread(self._main_loop, tuple())
        self._load_playlists()

    def close(self):
        self.downloader.stop()
        self.player.quit()

    def __repr__(self):
        return '<Player[%d] playing %s (views=%d)>'%(len(self.playlist), self.selected, len(self.views))

    def _main_loop(self):
        errors = dict(count=0)

        while True:
            if not self._paused and len(self.playlist):
                try:
                    try:
                        with self._lock:
                            p = self.player.position
                            if p is None:
                                errors['count'] = 10
                                self.position = None
                            else:
                                self.position = int(p)
                    except IOError, e:
                        web.debug('E: %s'%e)
                        self.position = None
                        # restart player
                        # self.player.respawn()
                    except:
                        DEBUG()
                        self.position = None

#                    web.debug('pos: %s, errors: %s'%(self.position, errors))

                    if self.position is None:
                        web.debug('NO POSITION %s'%errors)
                        if errors['count'] > 2:
                            errors['count'] = 0
                            i = self.select(1)
                        else:
                            errors['count'] += 1
                except Exception, e:
                    DEBUG(False)
            sleep(1)

    #@classmethod
    def _download_zic(self, uri, sync=False):
        d = media_config[get_type_from_uri(uri)]
        return self.downloader.get(uri, -1 if sync else d['init_chunk_size'], d['chunk_size'])

    def select(self, sense):
        """ Selects a song, according to the given offset
        ex.: self.select(1) # selects the next track
        self.select(-1) # selects the previous track
        self.select(0) # no-op
        """
        with self._lock:
            old_pos = self.playlist.pos
            try:
                self.playlist.move(sense)
            except EndOfPlaylist:
                if not config.loop:
                    self.pause()
            song_name = config.streaming_file
            if len(self.playlist) and self.playlist.pos < 0:
                self.playlist.pos = 0
            sel = self.selected
            uri = sel['uri']
            web.debug('download: %s'%uri)
            if uri.count('/db/get') != 1 or uri.count('id=') != 1:
                # something strange (LIVE mode)
                song_name = uri
                streaming = True
            else: # zicbee
                sync = bool(sel.get('cursed'))
                song_name = self._download_zic(sel['uri'], sync) # threaded
                streaming = False

            sibling = self.sibling
            if sibling and not sibling.get('cursed'):
                self.downloader.next = sibling['uri']

            web.debug('select: %d (previous=%s)'%(sel['pls_position'], old_pos))
            if old_pos != sel['pls_position']:
                sleep(0.3) # FIXME: wait for things to settle... :(((
                web.debug("Loadfile %d/%s : %s !!"%(sel['pls_position'], sel['pls_size'], song_name))
                cache = media_config[self.selected_type].get('player_cache')
                if cache:
                    self.player.set_cache(cache)
            web.debug('LOADING: %s'%song_name)
            self.player.load(song_name)
            description="""<b>%(title)s</b>
<i>%(album)s</i>"""%sel
            notify(sel.get('artist', 'Play'), description, icon='info') # generic notification
            self._paused = False
            if streaming:
                sleep(1.5) # FIXME: wait for things to settle... :(((
        return

    def volume(self, val):
        self.player.volume(val)

    def tag(self, tag):
        ci = self.selected['id']
        uri = 'http://%s/db/tag/%s/%s'%(self.hostname, ci, tag)
        urllib.urlopen(uri)

    def rate(self, score):
        web.debug('DEBUG: %s'%self.selected)
        ci = self.selected['id']
        uri = 'http://%s/db/rate/%s/%s'%(self.hostname, ci, score)
        web.debug('URI: %s'%uri)
        urllib.urlopen(uri)

    def shuffle(self):
        """ Shuffle the playlist, and selects the first track
        if the playlist is empty, do nothing
        """

        with self._lock:
            self.playlist.shuffle()
            notify('Shuffled', timeout=200, icon='shuffle')

    def seek(self, val):
        """ Seek according to given value
        """
        with self._lock:
            self.player.seek(val)
            notify('Seeking %s'%val, timeout=200, icon='next')

    def clear(self):
        """ Clear the current playlist and stop the player
        """
        with self._lock:
            self.playlist.clear()
            self._load_playlists()
            self.position = None
            self.player.respawn()
            self._paused = False

    def pause(self):
        """ (Un)Pause the player
        """
        with self._lock:
            self.player.pause()
            self._paused = not self._paused
        if self._paused:
            notify('Pause', timeout=300, icon='pause')
        else:
            notify('Play', timeout=300, icon='play')

    def delete_entry(self, position):
        """ delete the song at the given position """
        del self.playlist[position]

    def move_entry(self, pos1, pos2):
        """ Move an entry to a given playlist position"""
        p = self.playlist
        if p.pos == pos2:
            return
        elif p.pos > pos1:
            p.pos -= 1
            pos2 -= 1
        to_move = p[pos1]
        del p[pos1]
        p.inject(to_move, pos2)

    def swap_entry(self, pos1, pos2):
        """ Swap two entries in the current playlist """
        p = self.playlist
        p[pos1], p[pos2] = p[pos2], p[pos1]

    def playlist_change(self, operation, pls_name):
        """ copy or append a named playlist to the active one
        """
        pls = self._named_playlists[pls_name]
        if operation == 'copy':
            from copy import copy
            self.playlist = copy(pls)
            if config.autoshuffle:
                self.shuffle()
            self.select(0)
        elif operation == 'append':
            self.playlist.extend(pls)
        elif operation == 'inject':
            self.playlist.inject(pls, self.playlist.pos)
        self.running = True

    def delete_playlist(self, name):
        """ Delete the given named playlist (by name) """
        del self._named_playlists[name]
        self._save_playlists()

    def save(self, name):
        from copy import copy
        self._named_playlists[web.input()['name']] = copy(self.playlist)
        self._save_playlists()

    def fetch_playlist(self, hostname=None, playlist=False, **kw):
        """
        Fetch a playlist from a given hostname,
        can take any keyword, will be passed with the remote command
        kw should contain "pattern" parameter
        you can pass a playlist name to save the search
        a temporary playlist will be created with the given name
        (the main one is not affected)
        """
        self.running = False # do not disturb !
        web.debug('fetch_pl: h=%s tmp=%s %s'%(hostname, playlist, kw))

        # sanitise hostname
        hostname = hostname.strip()
        if not hostname:
            hostname = '127.0.0.1'
        if ':' not in hostname:
            hostname = "%s:%s"%(hostname, config.default_port)

        # set playlist name
        tokens = parse_string(kw['pattern'])
        for tok in tokens:
            if tok.isa(PLAYLIST):
                playlist = tok.value
                tokens.remove(tok)
                kw['pattern'] = tokens2string(tokens)
                break
        else:
            playlist = None


        with self._lock:
            self.hostname = hostname
            notify('Requesting', kw.get('pattern'))
            params = '&%s'%urllib.urlencode(kw) if kw else ''
            uri = 'http://%s/db/?fmt=json%s'%(hostname, params)
            site = urllib.urlopen(uri)

            # convert outpout playlist parameter
            append = False
            if playlist and playlist[0] in '>+':
                if playlist[0] == '>':
                    append = True
                else: # +
                    try:
                        append = self.playlist.pos + 1
                    except TypeError:
                        append = True
                playlist = playlist[1:].strip()

            elif not playlist:
                playlist = '#'

            if playlist == '#':
                out_pls = self.playlist
            else:
                out_pls = self._named_playlists[playlist] = Playlist()

            web.debug('fetch_pl: out_pls=%s kw=%s uri=%s'%(out_pls, kw, uri))

            # set up the output
            if not append:
                out_pls.replace([])

            if append and not isinstance(append, bool):
                def add_coroutine(offset):
                    n = offset
                    _injct = out_pls.inject
                    while True:
                        val = yield
                        _injct(val, n)
                        n = n+1
                iterator = add_coroutine(append)
                iterator.next()
                add = iterator.send
            else:
                add = out_pls.append

        total = 0

        while True:
            for n in xrange(50):
                line = site.readline()
                if not line:
                    break
                try:
                    r = jload(line)
                except:
                    DEBUG()
                    web.debug("Can't load json description: %s"%line)
                    break
                total += r[4]
#                web.debug('r=%s'%r)
                with self._lock:
                    add(r)
                self.signal_view('update_total', total)

            yield ''

            if not line:
                del add
                break

        if out_pls is self.playlist:
            if config.autoshuffle and not append:
                self.shuffle()
        else:
            self._save_playlists()


    def _save_playlists(self):
        try:
            save_file = file(os.path.join(DB_DIR, 'playlists.pk'), 'w')
            p = Pickler(save_file)
            p.dump(self._named_playlists)
        except Exception, e:
            web.debug('ERROR: save_playlists: %s'%repr(e))
        else:
            web.debug('playlists saved: %s'%self._named_playlists.keys())

    def _load_playlists(self):
        self._named_playlists = dict()
        db_path = os.path.join(DB_DIR, 'playlists.pk')
        try:
            save_file = file(db_path, 'r')
            p = Unpickler(save_file)
            self._named_playlists = p.load()
        except IOError, e:
            self._named_playlists['radios'] = Playlist((
                ['http://broadcast.infomaniak.net:80/radionova-high.mp3',
                    u'Misc artists', u'No album', 'Radio Nova', 1000,
                    None, None, 0],
                ['http://vipicecast.yacast.net:80/virginradio',
                    u'Misc artists', u'No album', 'Virgin Radio', 1000,
                    None, None, 0],
                ['http://vipicecast.yacast.net:80/vra_webradio03',
                    u'Misc artists', u'No album', 'Virgin Radio 70', 1000,
                    None, None, 0],
                ['http://streaming.rtbf.be:8000/2128xrtbf',
                    u'Misc artists', u'No album', 'Classic 21', 1000,
                    None, None, 0],
                ['http://ogg.frequence3.net:19000/frequence3.ogg',
                    u'Misc artists', u'No album', 'Radio Frequence3', 1000,
                    None, None, 0],
                ['http://mp3.live.tv-radio.com/lemouv/all/lemouvhautdebit.mp3',
                    u'Misc artists', u'No album', 'Radio Le Mouv\'', 1000,
                    None, None, 0],
                ['http://mp3.live.tv-radio.com/francemusique/all/francemusiquehautdebit.mp3',
                    u'Misc artists', u'No album', 'Radio France Musique', 1000,
                    None, None, 0],
                ))
            self._save_playlists()
        except Exception, e:
            web.debug('Unable to load playlist. Corrupted file ? (%s)'%repr(e))
            DEBUG()
        else:
            web.debug('playlists loaded: %s'%self._named_playlists.keys())

    def signal_view(self, name, *args, **kw):
#        web.debug('signaling %s %s %s'%(name, args, kw))
        for view in self.views:
            web.debug(' -> to view %s'%view)
            try:
                getattr(view, 'SIG_'+name)(*args, **kw)
            except:
                DEBUG()

    @property
    def selected(self):
        return self.playlist.selected_dict or None

    @property
    def sibling(self):
        return self.playlist.sibling


    @property
    def selected_type(self):
        # http://localhost:9090/db/get/song.mp3?id=5 -> mp3
        return get_type_from_uri(self.selected.get('uri'))




# vim: et ts=4 sw=4
from __future__ import with_statement

__all__ = ['web_db_index', 'WEB_FIELDS', 'render']

import os
import web
import random
from threading import RLock
from time import time
from zicbee.core import zshell
from zicbee_lib.formats import compact_int, jdump, uncompact_int, dump_data_as_text
from zicbee_lib.parser import string2python
from zicbee_lib.config import config
from zicbee_lib.debug import DEBUG
from zicbee_lib import debug
from zicbee import __version__ as VERSION

WEB_FIELDS = 'artist album title length score tags __id__'.split()

web.internalerror = web.debugerror = web.debug = debug.log.debug

if debug.debug_enabled:
    web.config.debug = True
else:
    web.config.debug = None

def kill_server():
    zshell.songs.close()
    try:
        from zicbee.core.httpplayer import webplayer
        webplayer.player.close()
    except:
        pass

def allow_admin_mode():
    addr = web.ctx.env['REMOTE_ADDR']

    admins = config.allow_remote_admin
    allowed = ['127.0.0.1']
    if admins:
        if len(admins) == 1 and admins[0].lower() in ('yes', 'true', 'on'):
            return True
        allowed.extend(admins)
    return addr in allowed


# Set default headers & go to templates directory
web.ctx.headers = [('Content-Type', 'text/html; charset=utf-8'), ('Expires', 'Thu, 01 Dec 1994 16:00:00 GMT')]
from zicbee_lib.resources import resource_filename
try:
    render = web.template.render(resource_filename('zicbee.ui.web', 'templates'))
except Exception, e:
    DEBUG()
    class FakeRender(object):
        def __getattr__(self, name):
            return self
        def __call__(self, *args, **kw):
            return 'Unable to load templates'
    render = FakeRender()

# DB part

available_formats = [
        ('html', 'WWW Browser'),
        ('txt', 'Text'),
        ('json', 'JSON'),
        ('m3u', 'Playlist (m3u)'),
        ('zip', 'ZIP archive'),
        ]

DbSimpleSearchForm = web.form.Form(
        web.form.Hidden('id'),
        web.form.Textbox('pattern', description="Play pattern"),
        web.form.Dropdown('fmt', available_formats,
            value="html", description="Output"),
        )

import zipfile
#import cZipFile


class Zipper(list):
    """ Object used to handle zip files """
    def __init__(self, zipname):
        self.name = zipname

#    def extract(self, name=None, to_path=None, members=None, load=False):
#        name = name or self.name
#        zf = zipfile.ZipFile(name)
#        if load:
#            self[:] = [(x, '') for x in zf.namelist()]
#        zf.extractall(to_path, members)
#        zf.close()

    def saveSongs(self, as_name=None, prefix=''):
        name = as_name or self.name
        zf = zipfile.ZipFile(name, 'w', zipfile.ZIP_DEFLATED)
        for uri, song in self:
            fname = song.filename
            zf.write(fname, "/".join((song.artist, song.album, song.title))+"."+(fname.rsplit('.', 1)[-1]))
        zf.close()

#    def saveFiles(self, as_name=None, prefix=''):
#        name = as_name or self.name
#        zf = zipfile.ZipFile(name, 'w', zipfile.ZIP_DEFLATED)
#        for fname, dirname in self:
#            zf.write(fname, prefix+fname[len(dirname):])
#        zf.close()


def refresh_db():
    zshell.songs.commit()
    zshell.songs._create()
    zshell.songs.cleanup()

class web_db_index:

    _db_lock = RLock()

    def REQ_tag(self, song, tag):
        """ Tags a song
        tag/<song id>/<tag>

        Args:
            - song (int): song
            - tag (str): the tag you want to add to that song
        Returns:
            None
        """

        song_id = uncompact_int(song)
        try:
            tag = unicode(tag)
            with self._db_lock:
                SEP=u','
                _t = zshell.songs[song_id].tags
                tag_set = set( _t.strip(SEP).split(SEP) ) if _t else set()
                for single_tag in tag.split(','):
                    tag_set.add(single_tag.strip())
                new_tag = ''.join((SEP,SEP.join(tag_set),SEP))
                zshell.songs[song_id].update(tags=new_tag)
        except Exception, e:
            web.debug('E!%s'%e)
        finally:
            refresh_db()

    def REQ_rate(self, song, rating):
        """ Rate a song
        rate/<song id>/<rating>

        Args:
            - song (int): song
            - rating (int): the rate you want to give to that song
        Returns:
            None
        """
        web.debug('rate: song=%s rating=%s'%(song,rating))
        try:
            with self._db_lock:
                song_id = uncompact_int(song)
                zshell.songs[song_id].update(score=int(rating))
        finally:
            refresh_db()

    def REQ_multirate(self, ratings_list):
        """ Rate a list of songs
        multirate/[song id=rating]<,song id2=rating2><,song id3=rating3>...

        Args:
            - ratings_list: a list of tuples of the form::
                song=rating,song2=rating2,...
        Returns:
            None
        """
        web.debug('rate: ratings_list=%s'%ratings_list)
        try:
            ratings = [rating.split('=') for rating in ratings_list.split(',')]
            with self._db_lock:
                for song, rating in ratings:
                        song_id = uncompact_int(song)
                        zshell.songs[song_id].update(score=int(rating))
        finally:
            refresh_db()

    def REQ_version(self):
        """ Returns version of zicbee used on that instance

        Returns:
            None
        """
        yield VERSION

    def REQ_random(self):
        """ Returns a pattern with a random artist query
        If "what" parameter is specified, allow to change
        the random attribute, currently available:
         - artist
         - album
        """
        inp = web.input()
        what = inp.get('what', 'artist')
        if what.startswith('album'):
            what = ['album', list(zshell.songs.albums)]
        else:
            what = ['artist', list(zshell.songs.artists)]

        what[1] = random.choice(what[1])
        return 'pattern=%s:%s'%tuple(what)

    def REQ_artists(self):
        """ List all the artists

        Keywords:
            fmt (str, optional): output format

        Returns:
            A list of artists in desired format
        """
        inp = web.input()
        for d in dump_data_as_text(zshell.songs.artists, inp.get('fmt', 'txt')):
            yield d

    def REQ_albums(self):
        """ List all the albums

        Keywords:
            fmt (str, optional): output format

        Returns:
            A list of albums in desired format
        """
        inp = web.input()
        for d in dump_data_as_text(zshell.songs.albums, inp.get('fmt', 'txt')):
            yield d

    def REQ_genres(self):
        """ List all the genres

        Keywords:
            fmt (str, optional): output format

        Returns:
            A list of genres in desired format
        """
        inp = web.input()
        for d in dump_data_as_text(zshell.songs.genres, inp.get('fmt', 'txt')):
            yield d

    def REQ_set(self):
        if not allow_admin_mode():
            return "Not allowed"
        i = web.input()
        k = i.get('k')
        v = i.get('v')
        if v is not None:
            config[k] = v
        v = config[k]
        return "%s = %s"%(k, ', '.join(v) if isinstance(v, (list, tuple)) else v)

    def REQ_kill(self):
        """ Kills the application

        Returns:
            "Aaaah!"
        """
        if allow_admin_mode():
            kill_server()
            yield 'Aaaah!'
            raise SystemExit()

    def REQ_infos(self, song_id):
        """ Show infos about a song
        Either positional parameter or ``id`` keyword has to be given

        Keywords:
            fmt (str): output format
            id (str): id of the song
        Args:
            song_id (int): index of the song
        Returns:
            all the metadatas in desired format
        """


        af = DbSimpleSearchForm()
        if af.validates():
            af.fill()
            song_id = uncompact_int(af['id'].value)
            fmt = af['fmt'].value

        song = zshell.songs[song_id]
        d = dict( (f, getattr(song, f)) for f in song.fields ) if song else {}
        return dump_data_as_text(d, fmt)

    def REQ_get(self, *args):
        af = DbSimpleSearchForm()
        if af.validates():
            af.fill()
            song_id = uncompact_int(af['id'].value)
        filename = zshell.songs[song_id].filename
        web.header('Content-Type', 'application/x-audio')
        web.header('Content-Disposition',
                'attachment; filename:%s'%filename.rsplit('/', 1)[-1], unique=True)

        CHUNK=1024
        in_fd = file(filename)
        web.header('Content-Length', str( os.fstat(in_fd.fileno()).st_size ) )
        yield ''

        while True:
            data = in_fd.read(CHUNK)
            if not data: break
            y = (yield data)
        in_fd.close()
        return

    def GET(self, name):
        hd = web.webapi.ctx.homedomain
        inp = web.input()
        t0 = time()

        # XXX: move this to special commands, using dedicated Form()s
        # m3u flag has to move to "fmt"
        args = None
        song_id = None
        handler = None # special action handler executed before listing
        af = DbSimpleSearchForm()
        if af.validates():
            af.fill()
            song_id = af['id'].value

        if song_id:
            args = [song_id]

        if name:
            # Special actions (1rs level path)
            try:
                name, path = name.split('/', 1)
            except ValueError:
                name = name
                path = None
            try:
                handler = getattr(self, 'REQ_' + name)
            except AttributeError:
                handler = None

            if not args:
                if path:
                    args = path.split('/')
                else:
                    args = []

        if handler:
            # XXX: replace that with autodelegate as in httpplayer
            try:
                # execute the handler
                web.debug('%s: %r'%(handler, args))
                ret = handler(*args)
                if hasattr(ret, 'next'):
                    for chunk in ret:
                        yield chunk
                else:
                    yield ret
            except GeneratorExit:
                raise
            except Exception, e:
                DEBUG()

        else: # XXX: move that to a dedicated command ? (ex: .../db/q?pattern=... looks nice)
            # or use "index" ... (sounds good too !)
            format = af['fmt'].value or 'html'
            if format == 'm3u':
                web.header('Content-Type', 'audio/x-mpegurl')
            elif format == 'html':
                web.header('Content-Type', 'text/html; charset=utf-8')

            pattern = af['pattern'].value

            if pattern is None:
                res = xrange(0)
            else:
                pat, kw = string2python(pattern)
                web.debug('PAT %r KW %r'%(pat, kw))
                urlencode = web.http.urlencode
                ci = compact_int

                res = ([hd+'/db/get/%s?id=%s'%('song.'+ r.filename.rsplit('.', 1)[-1].lower(), ci(int(r.__id__))), r]
                        for r in zshell.songs.search(list(WEB_FIELDS+['filename']), pat, **kw)
                        )
            t_sel = time()

            if format == 'm3u':
                yield unicode(render.m3u(web.http.url, res))
            elif format == 'zip':
                web.header('Content-Type', 'application/zip')
                import tempfile
                tmp_file = tempfile.NamedTemporaryFile()
                z = Zipper(tmp_file.name)
                z[:]= res
                z.saveSongs()
                tmp_file.seek(0)
                CZ = 2**18

                while True:
                    dat = tmp_file.read(CZ)
                    if not dat:
                        break
                    yield dat

                tmp_file.close()
                return
            elif not format or format == 'html':
                yield unicode(render.index(af, res, config.web_skin or 'default'))
            elif format in ('json', 'txt'):
                some_db = zshell.songs.databases.itervalues().next()['handle']
                # try to pre-compute useful things
                field_decoder = zip( WEB_FIELDS,
                        (some_db.f_decode[some_db.fields[fname]] for fname in WEB_FIELDS)
                        )
                yield ''

                infos_iterator = ( [r[0]] + [d(r[1][r[1].fields.index(f)]) for f, d in field_decoder]
                        for r in res )
                if format == 'json':
                    d = jdump
                else:
                    def d(line):
                        return ' | '.join(line[:4])

                try:
                    # TODO: optimise this (jdump by 100 size blocks)
                    for r in infos_iterator:
                        yield d(r)
                        yield '\n'
                    web.debug('handled in %.2fs (%.2f for select)'%(time() - t0, t_sel - t0))
                except Exception, e:
                    DEBUG()
                    web.debug("ERR: %s(%s)"% (type(e), e))
            else:
                # TODO: add support for zip output (returns a zip stream with all the songs)
                web.debug('unknown search mode')
                for r in res:
                    yield dump_data_as_text(r, format)




# vim: et ts=4 sw=4

import itertools
import socket
from zicbee_lib.config import config
from zicbee_lib.debug import log
from zicbee_lib.formats import get_help_from_func

try:
    socket.setdefaulttimeout(int(config.socket_timeout)) # setsocket  timeout, for interactive cmds
except Exception, e:
    log.warning("unable to set socket timeout to '%s': %s.", config.socket_timeout, e)

from .misc_commands import do_hash, do_reset, do_bundle, do_debug, do_cleanup
from .set import do_set
from .list import do_list
from .find_dups import do_find_dups
from .serve import do_serve
from .search import do_search
from .scan import do_scan

def do_help():
    """ Automatic help from self documentation """
    g = globals()
    undoc = []
    command_functions = [g[name] for name in g.keys() if name[:3] == 'do_']
    command_functions.sort()
    commands_display = []
    remote_commands_display = []
    for cmd in command_functions:
        cmd_help, cmd_is_remote = get_help_from_func(cmd)

        if cmd_is_remote:
            remote_commands_display.append(cmd_help)
        else:
            commands_display.append(cmd_help)

        if not '\n' in cmd_help:
            undoc.append(cmd.func_name[3:])

    if remote_commands_display:
        it = itertools.chain( ['[REMOTE COMMANDS]\n'], remote_commands_display, ['[LOCAL COMMANDS]\n'], commands_display )
    else:
        it = itertools.chain( ['[commands list]\n'], commands_display )
    for cmd in it:
        print cmd

    if undoc:
        print "Not documented:", ', '.join(undoc)




import zicbee_lib as zl

def notify(title=None, description=None, icon=None, timeout=None):
    zl.debug.log.info("[%s] %s: %s"%(icon, title, description))

try:
    if not zl.config.config.notify:
        raise ValueError('notify disabled in user settings')
    from .notify import notify
except ValueError, e:
    zl.debug.log.warning("Not loading notification: %s", e.args[0])
except Exception, e:
    zl.debug.log.error("Can't load notify framework! %s"%e)
    zl.debug.DEBUG(False)




from zicbee_lib.resources import resource_filename
import pynotify
import pygtk
pygtk.require('2.0')
import gtk

icons = 'connect refresh convert dialog-info dialog-error find media-pause media-next-ltr media-play-ltr media-next-rtl'.split()
shortnames = {
        'info': 'dialog-info',
        'error': 'dialog-error',
        'play': 'media-play-ltr',
        'pause': 'media-pause',
        'next': 'media-next-ltr',
        'prev': 'media-next-rtl',
        'previous': 'media-next-rtl',
        'shuffle': 'refresh', # can't find better...
        }
bee_icon = resource_filename('zicbee.ui.notify', 'bee_icon.png')

def notify(title, description=None, icon=None, timeout=750):
    if icon in shortnames:
        icon = shortnames[icon]

    if icon in icons:
        iname = "gtk-"+icon
    else:
        iname = 'file://%s'%bee_icon

    pynotify.init("SimpleNotifier")
    n = pynotify.Notification(title, description, iname)
    n.set_urgency(pynotify.URGENCY_LOW)
    n.set_timeout=(timeout)
    n.show()


if __name__ == '__main__':
    notify('Title1', 'descr1\nfoo\tbar\nblah\ttoto', icon='dialog-info')
#    notify('Title2', 'descr2', icon='media-play', timeout=200)
#    notify('Title3', 'descr3', '/tmp/iconbee.png')




# vim: et ts=4 sw=4
import os
from zicbee_lib.resources import set_proc_title
from zicbee.core.httpdb import kill_server
from zicbee_lib.config import config
from zicbee_lib.debug import DEBUG

def abort():
    kill_server()
    raise SystemExit()

def make_app(no_player=False, fork=None):
    import web
    import signal
    from zicbee_lib.debug import debug_enabled
    from zicbee_lib.resources import resource_filename
    from zicbee.core.httpdb import web_db_index

    # Handle ^C keystroke (try atexit if any problem...)
    signal.signal(signal.SIGINT, abort)

    if not no_player:
        # let's do webplayer
        try:
            from zicbee.core.httpplayer import webplayer
        except (ImportError, RuntimeError):
            print "Can't load webplayer, falling-back to pure db mode"
            DEBUG()
            no_player = True

    try:
        # chdir to serve files at the right place
        p = os.path.dirname(resource_filename('zicbee.ui.web', 'static'))
        os.chdir( p )
    except Exception:
        DEBUG()

    pid = 0 # if not forking, still execute children commands
    do_detach = False # do not try to detach by default

#    if config.fork and not debug_enabled or fork is True:
    if False: # FIXME: fork looks broken
        try:
            pid = os.fork()
            do_detach = True # fork succeded, try to detach
        except Exception, e:
            print "Can't fork: %s."%e

    if pid == 0:

        if do_detach:
            os.setsid()

        print "%s listening on:"%('Song browser' if no_player else 'Song browser and player'),
        if no_player:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'web_db_index'
                    )
        else:
            urls = ('/db/(.*)', 'web_db_index',
                    '/(.*)', 'webplayer')

        app = web.application(urls, locals())
        return app

def do_serve(pure=False):
    """ Create a ZicDB instance
    parameters:
        pure (default: False): just start DB serving, no player
    """
    import socket
    import sys

    set_proc_title('zicserve')
    sys.argv = ['zicdb', '0.0.0.0:%s'%(config.default_port)]

    try:
        make_app(no_player=pure).run()
    except SystemExit:
        print "ciao!"
    except socket.error:
        print "Already running!"
        try:
            abort()
        finally:
            raise SystemExit()
    except:
        DEBUG()
        try:
            abort()
        finally:
            print os.kill(os.getpid(), 9)
        #print 'kill', os.getpid()




import heapq
import itertools
from zicbee.core import zshell

def do_find_dups(wpt=None, ar=None):
    """
    Find duplicates (WIP)
    Parameters:
        wpt: wrong positive threshold (ceil to not reach), default == auto
        ar: auto remove (ask for directory deletion), the smallest directory always wins
    """
    hash_dict = dict()

    cnt = itertools.count()
    total_cnt = itertools.count()

    if wpt is None:
        wpt = min(1000, len(zshell.songs)/60) # take untaged/corrupted data into account

    for num, footprint in zshell.songs.get_hash_iterator():
        if footprint not in hash_dict:
            hash_dict[footprint] = [num]
        else:
            hash_dict[footprint].append(num)

    if ar:
        print '#!/bin/sh'
        for m in (matches for num, matches in hash_dict.iteritems() if 1 < len(matches) < wpt):
            h = []
            cnt.next()
            for num in m:
                total_cnt.next()
                song = zshell.songs[num]
                score = song.length + (len(song.track) if song.track else -10) + len(song.title) + len(song.genre) + len(song.artist) + len(song.album) + (-10 if '/32 ' in song.filename else 0)
                heapq.heappush(h,
                        (-score, song))

            for nb, other in h:
                ref = h[0][1]
                if other != ref:
                    if ref.length - 30 < other.length < ref.length + 30:
                        print "rm '%s'"%(other.filename.replace("'", r"\'"))
                    else:
                        print "# rm '%s'"%(other.filename.replace("'", r"\'"))
                else:
                    print "# kept %s"%other.filename
    else:
        for m in (matches for num, matches in hash_dict.iteritems() if 1 < len(matches) < wpt):
            print "#", cnt.next()
            for num in m:
                total_cnt.next()
                print "%d: %s"%(num, zshell.songs[num].filename)
    print "# %d songs are duplicates."%(total_cnt.next()-cnt.next())




from time import time
import sys
from zicbee.db import valid_tags
from zicbee.core import zshell
from zicbee_lib.parser import string2python
from zicbee_lib.formats import duration_tidy, jload
from zicbee_lib.config import config
from zicbee_lib.debug import log, DEBUG

def do_search(out=None, edit_mode=False):
    """ Search for song, display results.
    out can be "m3u" or "null", defaults to human-readable
    """

    duration = 0
    start_t = time()

    fields = list(valid_tags)
    fields.remove('filename')
    fields = tuple(fields)

    if callable(out):
        song_output = out
    elif out == 'm3u':
        print "#EXTM3U"
        def song_output(song):
            print u"#EXTINF:%d,%s - %s\n%s"%(song.length, song.artist, song.title, song.filename)
    elif out == 'null':
        def song_output(song): pass
    else:
        def song_output(song):
            txt = '%s :\n%s [%s, score: %s, tags: %s]'%(song.filename,
                    '%s - %s - %s'%(song.artist, song.album, song.title),
                    duration_tidy(song.length), song.score, song.tags,
                    )
            print txt.decode('utf8').encode('utf8')

    pat, kw = string2python(' '.join(zshell.args))
    if edit_mode:
        search_fn = zshell.songs.u_search
    else:
        search_fn = zshell.songs.search

    num = 0
    for num, res in enumerate(search_fn(None, pat, **kw)):
        song_output(res)
        duration += res.length

    sys.stderr.write("# %d results in %s for a total of %s!\n"%(
            num,
            duration_tidy(time()-start_t),
            duration_tidy(duration)
            ))





# vim: et ts=4 sw=4

def serve():
    startup('serve')

def setup_db(db_name, args):
    import zicbee.core.zshell as zshell
    zshell.init(args, db_name)

def parse_cmd(action='help', *arguments):
    db_name = None
    arguments = list(arguments)

    if action == 'use':
        db_name = arguments[0]
        action = arguments[1]
        del arguments[:2] # Remove "<db name> <action>"

    if '::' in action:
        params = action.split('::')
        action = params.pop(0)
        kparams = dict()
        for p in list(params):
            if '=' in p:
                k, v = p.split('=', 1)
                kparams[k] = v
                params.remove(p)
    else:
        kparams = dict()
        params = tuple()

    return db_name, arguments, action, params, kparams

def execute_cmd(action, *params, **kparams):
    try:
        import zicbee.core.commands as cmds
        commands_dict = dict((i[3:], getattr(cmds, i)) for i in dir(cmds) if i.startswith('do_'))
        commands_dict.get(action, cmds.do_help)(*params, **kparams)
    except KeyboardInterrupt:
        print "Abort!"

def shell():
    startup('shell')

def startup(*args):
    import os
    import sys
    db_name = None
    if not args:
        arguments = sys.argv[2:]
        if len(sys.argv) > 1:
            action = sys.argv[1]
        else:
            action = 'help'
        # feature: you can use aliases like zicdb-bigdrive to automatically use "bigdrive" db
        if '-' in os.path.basename(sys.argv[0]):
            db_name = sys.argv[0].rsplit('-', 1)[1]
    else:
        action = args[0]
        arguments = args[1:]

    cmd_db_name, arguments, action, params, kparams = parse_cmd(action, *arguments)
    setup_db(db_name or cmd_db_name, arguments)
    execute_cmd(action, *params, **kparams)




# vim: et ts=4 sw=4
__all__ = ['do_hash', 'do_reset', 'do_bundle', 'do_debug']

import urllib
from zicbee.core import zshell
from zicbee_lib.config import config, defaults_dict, DB_DIR
from zicbee.core.zshell import DEFAULT_NAME
from zicbee.db import Database, DB_DIR

def do_debug():
    """ Start a PDB (dev/hackers only)"""
    import pdb; pdb.set_trace()

def do_bundle():
    """ Dump used database to specified archive (any filename) """
    if len(zshell.args) != 1:
        raise SystemExit("Need filename name as agment !")
    zshell.songs.dump_archive(zshell.args[0])

def do_reset():
    """ Destroy all songs on used database """
    zshell.songs.destroy()
    print "Database cleared!"

def do_hash():
    """ List all songs by id and hash (mostly to debug find_dups command) """
    for i in zshell.songs.get_hash_iterator():
        print "%8d / %s"%i

def do_cleanup():
    """ Clean the database up """
    zshell.songs.cleanup()

def _webget(uri):
    if not uri.startswith('http://'):
        uri = "http://" + uri.lstrip('/')
    try:
        return urllib.urlopen(uri).read()
    except IOError, e:
        print "webget(%s): %s"%(uri, e)





# vim: et ts=4 sw=4
from zicbee.core import zshell
from zicbee_lib.config import config

def do_set():
    """ set a config variable to the given value
    list all variables and values if no argument is given"""

    if not zshell.args:
        # dumps *
#        print "[DEFAULT]"

        not_found = []
        for k, v in config:
            print "%s = %s"%(k, v)

    elif zshell.args and len(zshell.args) < 4:
        # (re)set a value

        if len(zshell.args) == 2 or (len(zshell.args) == 3 and zshell.args[1] == '='):
            # set a value
            if zshell.args[1] == '=':
                del zshell.args[1]
            setattr(config, zshell.args[0], zshell.args[1])
        elif len(zshell.args) == 1:
            # reset (blank) a value
            setattr(config, zshell.args[0], '')

    else:
        print """Takes exactly 2 arguments: set <variable> <value>, takes no param to list variables.
        giving no arguments will print every variable
        a single parameter will be reset (blanked) """



# vim: et ts=4 sw=4
from __future__ import with_statement

import difflib
import urllib
from zicbee.core.httpdb import render, web, allow_admin_mode
from zicbee.core.player import PlayerCtl
from zicbee.utils import notify
from zicbee_lib.config import config
from zicbee_lib.debug import DEBUG
from zicbee_lib.formats import dump_data_as_text, jdump, get_index_or_slice

SimpleSearchForm = web.form.Form(
        web.form.Hidden('id'),
        web.form.Textbox('host', description='Search host', value='localhost'),
        web.form.Textbox('pattern', description='Search pattern'),
#        web.form.Textbox('tempname', description='Temporary name'),
        )

TagForm = web.form.Form(web.form.Textbox('tag', description='Set tag'))
ScoreForm = web.form.Form(web.form.Dropdown('score', range(11), description='Set rate'))

class webplayer:
    player = PlayerCtl()

    GET = web.autodelegate('REQ_')

    def REQ_main(self):
        return self.render_main(render.player)

    def REQ_basic(self):
        return self.render_main(render.basicplayer)

    def render_main(self, rend):
        cook_jar = web.cookies(host='localhost', pattern='')
        cook_jar['pattern'] = urllib.unquote(cook_jar['pattern'])
        af = SimpleSearchForm(True)
        sf = ScoreForm(True)
        tf = TagForm(True)
        af.fill(cook_jar)
        yield unicode(rend(af, sf, tf, config.web_skin or 'default'))

    REQ_ = REQ_main # default page

    def REQ_close(self):
        if allow_admin_mode():
            web.redirect('/db/kill')

    def REQ_search(self):
        it = ('' for i in xrange(1))
        try:
            i = web.input()
            tempname = False

            if i.get('pattern', '').startswith('http'):
                # http pattern
                try:
                    uri = i.pattern.split()[0]
                    hostname = uri.split("/", 3)[2]
                    song_id = uri.rsplit('=', 1)[1]
                    it = self.player.fetch_playlist(hostname, pattern=u'id: %s pls: +#'%song_id)
                except:
                    # hardcore injection
                    pls = self.player.playlist

                    pls.inject( [str(uri), u'No artist', u'No album', 'External stream', 1000, None, None, 0] )
            else:
                # standard pattern
                it = self.player.fetch_playlist(i.get('host', 'localhost'), pattern=i.pattern)
            it.next()

        except (IndexError, KeyError):
            DEBUG(False)
        except Exception, e:
            DEBUG()
        finally:
            return it

    def REQ_delete(self):
        i = web.input()
        try:
            i = get_index_or_slice(i['idx'])
        except ValueError:
            self.player.delete_playlist(i['idx'])
        else:
            self.player.delete_entry(i)

        return ''

    def REQ_move(self):
        i = web.input()
        start = get_index_or_slice(i['s'])
        self.player.move_entry(start, int(i['d']))
        return ''

    def REQ_swap(self):
        i = web.input()
        self.player.swap_entry(int(i['i1']), int(i['i2']))
        return ''

    def REQ_append(self):
        self.player.playlist_change('append', web.input()['name'])
        return ''

    def REQ_copy(self):
        try:
            self.player.playlist_change('copy', web.input()['name'])
            return ''
        except KeyError:
            return 'Not Found'

    def REQ_inject(self):
        try:
            self.player.playlist_change('inject', web.input()['name'])
            return ''
        except KeyError:
            return 'Not Found'

    def REQ_save(self):
        name = web.input()['name']
        self.player.save(name)
        return 'saved %s'%name

    def REQ_volume(self):
        i = web.input()
        val = i.get('val')
        if val is not None:
            self.player.volume(val)
        return ''

    def REQ_infos(self):
        format = web.input().get('fmt', 'txt')

        _d = self.player.selected or dict()
        # add player infos
        _d['song_position'] = self.player.position
        _d['paused'] = self.player._paused

        if format.startswith('htm'):
            web.header('Content-Type', 'text/html; charset=utf-8')
        return dump_data_as_text(_d, format)

    def REQ_playlists(self):
        return '\n'.join(self.player._named_playlists.keys())

    def REQ_playlist(self):
        i = web.input()
        pls = self.player.playlist

        start = int(i.get('start', 0))

        format = i.get('fmt', 'txt')
        if start < 0:
            start = len(pls) + start

        if i.get('res'):
            r = int(i.res)
            if r > 0:
                end = start + r
            else:
                # compute from end
                end = len(pls) + r
        else:
            end = len(pls)

        window_iterator = (pls[i] + [i] for i in xrange(start, min(len(pls), end)))

        return dump_data_as_text(window_iterator, format)

    def REQ_guess(self, guess):
        try:
            self.player.selected.iteritems
        except AttributeError:
            yield jdump(False)
            return

        artist = self.player.selected['artist']
        title = self.player.selected['title']
        if difflib.get_close_matches(guess, (artist, title)):
            yield jdump(True)
        else:
            yield jdump(False)

    def REQ_shuffle(self):
        return self.player.shuffle() or ''

    def REQ_clear(self):
        return self.player.clear() or ''

    def REQ_pause(self):
        return self.player.pause() or ''

    def REQ_prev(self):
        notify('Zap!', icon='prev', timeout=200)
        return self.player.select(-1) or ''

    def REQ_next(self):
        notify('Zap!', icon='next', timeout=200)
        return self.player.select(1) or ''

    def REQ_tag(self, tag):
        return self.player.tag(unicode(tag.lstrip('/'))) or ''

    def REQ_rate(self, score):
        return self.player.rate(score.lstrip('/')) or ''

    def REQ_seek(self, val):
        val = val[1:]
        web.debug('VAL=%s'%val)
        self.player.seek(int(val))
        return ''




#!/usr/bin/env python
import os
import sys
try:
	import setuptools
except ImportError:
	from distribute_setup import use_setuptools
	use_setuptools()
from setuptools import setup, find_packages

sys.path.insert(0, '.')
import zicbee
VERSION=zicbee.__version__

if 'install' in sys.argv:
    print """Warning:
You will need to install some parts manually:

If it can't build, try to comment the line:
            'python-cjson>=1.0.5',

and uncomment one of those (recommended alternative: simplejson):
#            'simplejson>=1.7.3',
#            'demjson>=1.1',

Good luck !"""



# also supported:
#            'simplejson>=1.7.3',

requirements = [ 'zicbee-lib>=0.7.1', 'buzhug>=1.8', 'mutagen>=1.20', 'web.py>=0.34' ]

if sys.version_info[:2] < (2, 6):
    # add cjson dependency
    if os.name in ('nt', 'ce'):
        requirements.append( 'demjson>=1.1' )
    else:
        requirements.append( 'python-cjson>=1.0.5' )

DESCRIPTION=open('zicbee.rst').read()

setup (
        name='zicbee',
        version=VERSION,
        author='Fabien Devaux',
        author_email='fdev31@gmail.com',
        url = 'http://zicbee.gnux.info/',
        download_url='http://zicbee.gnux.info/hg/index.cgi/zicbee/archive/%s.tar.bz2'%VERSION,
        license='BSD',
        platform='all',
        description='A simple & powerful distributed Music database engine',
        long_description=DESCRIPTION,
        keywords = 'database music tags metadata management',
        include_package_data = True,
        packages = find_packages(),
        zip_safe = False,

        entry_points = {
            "console_scripts": [
                'zicdb = zicbee.core:startup',
                'zicserve = zicbee.core:serve',
                ],
            "setuptools.installation" : [
                'eggsecutable = zicbee.core:startup'
                ]
            },

        install_requires = requirements,

        extras_require = dict(
            player='zicbee-mplayer>=0.9',
            ),

        dependency_links = [
            'http://webpy.org/',
            'http://buzhug.sourceforge.net/',
            'http://webpy.org/static/',
            'http://code.google.com/p/quodlibet/downloads/list',
            ],
        classifiers = [
                'Development Status :: 4 - Beta',
                'Intended Audience :: Developers',
#                'Intended Audience :: End Users/Desktop',
                'Operating System :: OS Independent',
                'Operating System :: Microsoft :: Windows',
                'Operating System :: POSIX',
                'Programming Language :: Python',
                'Environment :: Console',
                'Environment :: No Input/Output (Daemon)',
                'Environment :: X11 Applications',
                'Natural Language :: English',
                'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
                'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
                'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
                'Topic :: Software Development',
                'Topic :: Software Development :: Libraries :: Python Modules',
                'Topic :: Multimedia :: Sound/Audio :: Players',
                'Topic :: Multimedia :: Sound/Audio :: Players :: MP3',
                'Topic :: Text Processing :: Markup',
                'Topic :: Utilities',
                ],

        )




# vim: et ts=4 sw=4
__all__ = ['songs', 'args', 'init']
import os
import sys
from zicbee.db import Database

DEFAULT_NAME='songs'
args = []

def init(args=None, db_name=None):
    try:
        db
    except NameError:
        pass
    else:
        print "db cleanup"
        db.cleanup() # XXX: Ugly !
        db.close()

    db_name = db_name or os.environ.get('ZDB', DEFAULT_NAME)
    print "opening %s..."%db_name
    db = Database(db_name)
    globals().update( dict(songs=db, args=args) )




# vim: et ts=4 sw=4
from zicbee_lib.config import DB_DIR
from zicbee.core.zshell import DEFAULT_NAME
from zicbee.db import Database, DB_DIR

def do_list():
    """ List available databases (some can be specified with "use" argument) """
    from os import listdir
    from os.path import isfile, join

    for i in listdir(DB_DIR):
        if isfile(join(DB_DIR, i, '__info__')) and isfile(join(DB_DIR, i, 'album')):
            txt = "%s # %d records"%(i, len(Database(i)))
            if i == DEFAULT_NAME:
                txt += ' [default]'
            print txt

